@@ -4,9 +4,9 @@ use anyhow::Result;
 use csv::{Position, Reader, ReaderBuilder};
 use std::cmp::max;
 use std::fs::File;
-use std::io::{BufRead, BufReader};
 use std::sync::{Arc, Mutex};
 use std::thread::{self, JoinHandle};
+use std::time;
 
 fn string_record_to_vec(record: &csv::StringRecord) -> Vec<String> {
     let mut string_vec = Vec::new();
@@ -143,13 +143,13 @@ impl CsvLensReader {
         Ok(reader)
     }
 
-    pub fn get_rows(&mut self, rows_from: u64, num_rows: u64) -> Result<Vec<Row>> {
+    pub fn get_rows(&mut self, rows_from: u64, num_rows: u64) -> Result<(Vec<Row>, GetRowsStats)> {
         let indices: Vec<u64> = (rows_from..rows_from + num_rows).collect();
-        self.get_rows_impl(&indices).map(|x| x.0)
+        self.get_rows_impl(&indices)
     }
 
-    pub fn get_rows_for_indices(&mut self, indices: &[u64]) -> Result<Vec<Row>> {
-        self.get_rows_impl(indices).map(|x| x.0)
+    pub fn get_rows_for_indices(&mut self, indices: &[u64]) -> Result<(Vec<Row>, GetRowsStats)> {
+        self.get_rows_impl(indices)
     }
 
     fn get_rows_impl(&mut self, indices: &[u64]) -> Result<(Vec<Row>, GetRowsStats)> {
@@ -175,7 +175,11 @@ impl CsvLensReader {
         let pos = Position::new();
         self.reader.seek(pos)?;
 
+        let tic = time::Instant::now();
         let pos_table = self.get_pos_table();
+        stats.pos_table_elapsed = Some(tic.elapsed());
+        stats.pos_table_entry = pos_table.len();
+
         let mut pos_iter = pos_table.iter();
         let mut indices_iter = indices.iter();
 
@@ -190,15 +194,19 @@ impl CsvLensReader {
             }
             // seek as close to the next wanted record index as possible
             let index = next_wanted.unwrap();
+            let mut seek_pos: Option<Position> = None;
             while let Some(pos) = next_pos {
                 if self.config.position_to_record_index(pos.record()) <= index.record_index {
-                    self.reader.seek(pos.clone())?;
-                    stats.log_seek();
+                    seek_pos.replace(pos.clone());
                 } else {
                     break;
                 }
                 next_pos = pos_iter.next();
             }
+            if let Some(pos) = seek_pos {
+                self.reader.seek(pos)?;
+                stats.log_seek();
+            }
 
             // note that records() excludes header by default, but here the first entry is header
             // because of the seek() above.
@@ -269,8 +277,14 @@ impl CsvLensReader {
         res
     }
 
-    pub fn get_total_line_numbers_approx(&self) -> Option<usize> {
-        let res = self.internal.lock().unwrap().total_line_number_approx;
+    pub fn get_last_indexed_line_number(&self) -> Option<usize> {
+        let res = self
+            .internal
+            .lock()
+            .unwrap()
+            .pos_table
+            .last()
+            .map(|x| x.record() as usize);
         res
     }
 
@@ -280,10 +294,12 @@ impl CsvLensReader {
     }
 }
 
-#[derive(Debug, PartialEq)]
-struct GetRowsStats {
-    num_seek: u64,
-    num_parsed_record: u64,
+#[derive(Debug, Clone, PartialEq)]
+pub struct GetRowsStats {
+    pub num_seek: u64,
+    pub num_parsed_record: u64,
+    pub pos_table_elapsed: Option<time::Duration>,
+    pub pos_table_entry: usize,
 }
 
 impl GetRowsStats {
@@ -291,6 +307,8 @@ impl GetRowsStats {
         GetRowsStats {
             num_seek: 0,
             num_parsed_record: 0,
+            pos_table_elapsed: None,
+            pos_table_entry: 0,
         }
     }
 
@@ -305,7 +323,6 @@ impl GetRowsStats {
 
 struct ReaderInternalState {
     total_line_number: Option<usize>,
-    total_line_number_approx: Option<usize>,
     pos_table: Vec<Position>,
     done: bool,
 }
@@ -314,7 +331,6 @@ impl ReaderInternalState {
     fn init_internal(config: Arc<CsvConfig>) -> (Arc<Mutex<ReaderInternalState>>, JoinHandle<()>) {
         let internal = ReaderInternalState {
             total_line_number: None,
-            total_line_number_approx: None,
             pos_table: vec![],
             done: false,
         };
@@ -323,31 +339,20 @@ impl ReaderInternalState {
 
         let _m = m_state.clone();
         let handle = thread::spawn(move || {
-            // quick line count
-            let mut total_line_number_approx;
-            {
-                let file = File::open(config.filename()).unwrap();
-                let buf_reader = BufReader::new(file);
-                // subtract 1 for headers
-                total_line_number_approx = buf_reader.lines().count();
-                if config.has_headers() {
-                    total_line_number_approx -= 1;
-                }
-
-                let mut m = _m.lock().unwrap();
-                m.total_line_number_approx = Some(total_line_number_approx);
-            }
-
+            let filesize = File::open(config.filename())
+                .unwrap()
+                .metadata()
+                .unwrap()
+                .len();
             let pos_table_num_entries = 10000;
-            let minimum_interval = 100; // handle small csv (don't keep pos every line)
-            let pos_table_update_every = max(
-                minimum_interval,
-                total_line_number_approx / pos_table_num_entries,
-            );
+            let minimum_interval = 500; // handle small csv (don't keep pos every byte)
+            let pos_table_update_every = max(minimum_interval, filesize / pos_table_num_entries);
 
             // full csv parsing
             let bg_reader = config.new_reader().unwrap();
-            let mut n = 0;
+            let mut n_lines = 0;
+            let mut n_bytes: u64 = 0;
+            let mut last_updated_at = 0;
             let mut iter = bg_reader.into_records();
             loop {
                 let next_pos = iter.reader().position().clone();
@@ -355,14 +360,17 @@ impl ReaderInternalState {
                     break;
                 }
                 // must not include headers position here (n > 0)
-                if n > 0 && n % pos_table_update_every == 0 {
+                let cur = n_bytes / pos_table_update_every;
+                if n_bytes > 0 && cur > last_updated_at {
                     let mut m = _m.lock().unwrap();
-                    m.pos_table.push(next_pos);
+                    m.pos_table.push(next_pos.clone());
+                    last_updated_at = cur;
                 }
-                n += 1;
+                n_lines += 1;
+                n_bytes = next_pos.byte();
             }
             let mut m = _m.lock().unwrap();
-            m.total_line_number = Some(n);
+            m.total_line_number = Some(n_lines);
             m.done = true;
         });
 
@@ -401,7 +409,7 @@ mod tests {
         let config = Arc::new(CsvConfig::new("tests/data/cities.csv", b',', false));
         let mut r = CsvLensReader::new(config).unwrap();
         r.wait_internal();
-        let rows = r.get_rows(2, 3).unwrap();
+        let rows = r.get_rows(2, 3).unwrap().0;
         let expected = vec![
             Row::new(
                 3,
@@ -448,7 +456,7 @@ mod tests {
         let config = Arc::new(CsvConfig::new("tests/data/simple.csv", b',', false));
         let mut r = CsvLensReader::new(config).unwrap();
         r.wait_internal();
-        let rows = r.get_rows(1234, 2).unwrap();
+        let rows = r.get_rows(1234, 2).unwrap().0;
         let expected = vec![
             Row::new(1235, vec!["A1235", "B1235"]),
             Row::new(1236, vec!["A1236", "B1236"]),
@@ -472,7 +480,7 @@ mod tests {
         let mut r = CsvLensReader::new(config).unwrap();
         r.wait_internal();
         let indices = vec![1, 3, 5, 1234, 2345, 3456, 4999];
-        let (rows, stats) = r.get_rows_impl(&indices).unwrap();
+        let (rows, mut stats) = r.get_rows_impl(&indices).unwrap();
         let expected = vec![
             Row::new(2, vec!["A2", "B2"]),
             Row::new(4, vec!["A4", "B4"]),
@@ -483,9 +491,12 @@ mod tests {
             Row::new(5000, vec!["A5000", "B5000"]),
         ];
         assert_eq!(rows, expected);
+        stats.pos_table_elapsed.take();
         let expected = GetRowsStats {
-            num_seek: 49,
-            num_parsed_record: 505,
+            num_seek: 4,
+            num_parsed_record: 218,
+            pos_table_elapsed: None,
+            pos_table_entry: 115,
         };
         assert_eq!(stats, expected);
     }
@@ -496,12 +507,15 @@ mod tests {
         let mut r = CsvLensReader::new(config).unwrap();
         r.wait_internal();
         let indices = vec![1234];
-        let (rows, stats) = r.get_rows_impl(&indices).unwrap();
+        let (rows, mut stats) = r.get_rows_impl(&indices).unwrap();
         let expected = vec![Row::new(1235, vec!["A1235", "B1235"])];
         assert_eq!(rows, expected);
+        stats.pos_table_elapsed.take();
         let expected = GetRowsStats {
-            num_seek: 12,
-            num_parsed_record: 35,
+            num_seek: 1,
+            num_parsed_record: 8,
+            pos_table_elapsed: None,
+            pos_table_entry: 115,
         };
         assert_eq!(stats, expected);
     }
@@ -512,12 +526,15 @@ mod tests {
         let mut r = CsvLensReader::new(config).unwrap();
         r.wait_internal();
         let indices = vec![2];
-        let (rows, stats) = r.get_rows_impl(&indices).unwrap();
+        let (rows, mut stats) = r.get_rows_impl(&indices).unwrap();
         let expected = vec![Row::new(3, vec!["A3", "B3"])];
         assert_eq!(rows, expected);
+        stats.pos_table_elapsed.take();
         let expected = GetRowsStats {
             num_seek: 0,
             num_parsed_record: 4, // 3 + 1 (including header)
+            pos_table_elapsed: None,
+            pos_table_entry: 115,
         };
         assert_eq!(stats, expected);
     }
@@ -526,7 +543,7 @@ mod tests {
     fn test_small() {
         let config = Arc::new(CsvConfig::new("tests/data/small.csv", b',', false));
         let mut r = CsvLensReader::new(config).unwrap();
-        let rows = r.get_rows(0, 50).unwrap();
+        let rows = r.get_rows(0, 50).unwrap().0;
         let expected = vec![
             Row::new(1, vec!["c1", " v1"]),
             Row::new(2, vec!["c2", " v2"]),
@@ -538,7 +555,7 @@ mod tests {
     fn test_small_delimiter() {
         let config = Arc::new(CsvConfig::new("tests/data/small.bsv", b'|', false));
         let mut r = CsvLensReader::new(config).unwrap();
-        let rows = r.get_rows(0, 50).unwrap();
+        let rows = r.get_rows(0, 50).unwrap().0;
         let expected = vec![Row::new(1, vec!["c1", "v1"]), Row::new(2, vec!["c2", "v2"])];
         assert_eq!(rows, expected);
     }
@@ -547,7 +564,7 @@ mod tests {
     fn test_irregular() {
         let config = Arc::new(CsvConfig::new("tests/data/irregular.csv", b',', false));
         let mut r = CsvLensReader::new(config).unwrap();
-        let rows = r.get_rows(0, 50).unwrap();
+        let rows = r.get_rows(0, 50).unwrap().0;
         let expected = vec![Row::new(1, vec!["c1"]), Row::new(2, vec!["c2", " v2"])];
         assert_eq!(rows, expected);
     }
@@ -560,7 +577,7 @@ mod tests {
             false,
         ));
         let mut r = CsvLensReader::new(config).unwrap();
-        let rows = r.get_rows(0, 50).unwrap();
+        let rows = r.get_rows(0, 50).unwrap().0;
         let expected = vec![
             Row::new(1, vec!["1", "quote"]),
             Row::new(2, vec!["5", "Comma, comma"]),
@@ -573,7 +590,7 @@ mod tests {
         let config = Arc::new(CsvConfig::new("tests/data/simple.csv", b',', false));
         let mut r = CsvLensReader::new(config).unwrap();
         r.wait_internal();
-        let rows = r.get_rows_for_indices(&vec![1235, 1234]).unwrap();
+        let rows = r.get_rows_for_indices(&vec![1235, 1234]).unwrap().0;
         let expected = vec![
             Row::new(1236, vec!["A1236", "B1236"]),
             Row::new(1235, vec!["A1235", "B1235"]),
