@@ -1039,7 +1039,6 @@ pub struct CsvTableState {
     pub transient_message: Option<String>,
     pub echo_column: Option<String>,
     pub ignore_case: bool,
-    pub no_headers: bool,
     pub view_layout: Option<ViewLayout>,
     pub enable_line_wrap: bool,
     pub is_word_wrap: bool,
@@ -1054,7 +1053,6 @@ impl CsvTableState {
         total_cols: usize,
         echo_column: &Option<String>,
         ignore_case: bool,
-        no_headers: bool,
     ) -> Self {
         Self {
             rows_offset: 0,
@@ -1075,7 +1073,6 @@ impl CsvTableState {
             transient_message: None,
             echo_column: echo_column.clone(),
             ignore_case,
-            no_headers,
             view_layout: None,
             enable_line_wrap: false,
             is_word_wrap: false,
