@@ -7,18 +7,23 @@ use anyhow::Result;
 use regex::Regex;
 use std::cmp::min;
 use std::sync::Arc;
-use std::time::Instant;
+use std::time::{Duration, Instant};
 
 struct RowsFilter {
     indices: Vec<u64>,
     total: usize,
+    max_index: Option<u64>,
 }
 
 impl RowsFilter {
     fn new(finder: &find::Finder, rows_from: u64, num_rows: u64) -> RowsFilter {
-        let total = finder.count();
+        let (total, max_index) = finder.count_and_max_row_index();
         let indices = finder.get_subset_found(rows_from as usize, num_rows as usize);
-        RowsFilter { indices, total }
+        RowsFilter {
+            indices,
+            total,
+            max_index,
+        }
     }
 }
 
@@ -243,6 +248,12 @@ pub struct Header {
     pub origin_index: usize,
 }
 
+#[derive(Debug, Clone)]
+pub struct PerfStats {
+    pub elapsed: Duration,
+    pub reader_stats: crate::csv::GetRowsStats,
+}
+
 pub struct RowsView {
     reader: CsvLensReader,
     rows: Vec<Row>,
@@ -255,13 +266,13 @@ pub struct RowsView {
     columns_filter: Option<ColumnsFilter>,
     sorter: Option<Arc<Sorter>>,
     pub selection: Selection,
-    elapsed: Option<u128>,
+    perf_stats: Option<PerfStats>,
 }
 
 impl RowsView {
     pub fn new(mut reader: CsvLensReader, num_rows: u64) -> Result<RowsView> {
         let rows_from = 0;
-        let rows = reader.get_rows(rows_from, num_rows)?;
+        let rows = reader.get_rows(rows_from, num_rows)?.0;
         let headers = Self::get_default_headers_from_reader(&reader);
         let view = Self {
             reader,
@@ -275,7 +286,7 @@ impl RowsView {
             columns_filter: None,
             sorter: None,
             selection: Selection::default(num_rows),
-            elapsed: None,
+            perf_stats: None,
         };
         Ok(view)
     }
@@ -474,8 +485,8 @@ impl RowsView {
             .map(|x| x.saturating_add(self.rows_from))
     }
 
-    pub fn elapsed(&self) -> Option<u128> {
-        self.elapsed
+    pub fn perf_stats(&self) -> Option<PerfStats> {
+        self.perf_stats.as_ref().cloned()
     }
 
     pub fn get_total_line_numbers(&self) -> Option<usize> {
@@ -483,7 +494,7 @@ impl RowsView {
     }
 
     pub fn get_total_line_numbers_approx(&self) -> Option<usize> {
-        self.reader.get_total_line_numbers_approx()
+        self.reader.get_last_indexed_line_number()
     }
 
     pub fn in_view(&self, row_index: u64) -> bool {
@@ -539,7 +550,7 @@ impl RowsView {
                 self.selection.row.select_first()
             }
             Control::ScrollBottom => {
-                if let Some(total) = self.get_total() {
+                if let Some(total) = self.get_total_line_numbers_indexed() {
                     // Note: Using num_rows_rendered is not exactly correct, but it's simple and
                     // a bit better than num_rows. To be exact, this should use row heights to
                     // determine exactly how many rows to show from the bottom.
@@ -561,15 +572,24 @@ impl RowsView {
         Ok(())
     }
 
-    fn get_total(&self) -> Option<usize> {
-        if let Some(filter) = &self.filter {
-            return Some(filter.total);
-        } else if let Some(n) = self
+    fn get_total_line_numbers_indexed(&self) -> Option<usize> {
+        if let Some(max_line_number) = self
             .reader
             .get_total_line_numbers()
-            .or_else(|| self.reader.get_total_line_numbers_approx())
+            .or_else(|| self.reader.get_last_indexed_line_number())
         {
-            return Some(n);
+            if let Some(filter) = &self.filter {
+                if let Some(max_index) = filter.max_index {
+                    if max_index < max_line_number as u64 {
+                        // Only allow jumping to the bottom of found records if it can be
+                        // efficiently retrieved
+                        return Some(filter.total);
+                    }
+                    return None;
+                }
+            } else {
+                return Some(max_line_number);
+            }
         }
         None
     }
@@ -588,7 +608,7 @@ impl RowsView {
 
     fn bottom_rows_from(&self) -> Option<u64> {
         // fix type conversion craziness
-        if let Some(n) = self.get_total() {
+        if let Some(n) = self.get_total_line_numbers_indexed() {
             return Some(n.saturating_sub(self.num_rows_rendered as usize) as u64);
         }
         None
@@ -604,7 +624,7 @@ impl RowsView {
 
     fn do_get_rows(&mut self) -> Result<()> {
         let start = Instant::now();
-        let mut rows = if let Some(filter) = &self.filter {
+        let (mut rows, reader_stats) = if let Some(filter) = &self.filter {
             let indices = &filter.indices;
             self.reader.get_rows_for_indices(indices)?
         } else if let Some(sorter) = &self.sorter {
@@ -616,12 +636,15 @@ impl RowsView {
         } else {
             self.reader.get_rows(self.rows_from, self.num_rows)?
         };
-        let elapsed = start.elapsed().as_micros();
+        let elapsed = start.elapsed();
         if let Some(columns_filter) = &self.columns_filter {
             rows = Self::subset_columns(&rows, columns_filter.indices());
         }
         self.rows = rows;
-        self.elapsed = Some(elapsed);
+        self.perf_stats = Some(PerfStats {
+            elapsed,
+            reader_stats,
+        });
         // current selected might be out of range, reset it
         // self.selection.row.set_bound(self.rows.len() as u64);
         Ok(())
